import tensorflow as tf
import tensorflow.mlir.experimental as tme

print('\n#1')
x = tf.constant([[0.,1.],[2.,3.]], dtype=tf.float32);print(f"{x=}")
w = tf.constant([[1,0],[0,1]], dtype=tf.int32);print(f"{w=}")
y = tf.gather_nd(x, w,0); print(f"{y=}")
f =  tme.convert_function((tf.function(tf.gather_nd)).get_concrete_function(x, w, 0))
print(tme.run_pass_pipeline(f, "tf-lower-to-mlprogram-and-hlo"))


print('\n#2')
x = tf.constant([[0.,1.],[2.,3.]], dtype=tf.float32);print(f"{x=}")
w = tf.constant([[1],[0]], dtype=tf.int32);print(f"{w=}")
y = tf.gather_nd(x, w, 0); print(f"{y=}")
f =  tme.convert_function((tf.function(tf.gather_nd)).get_concrete_function(x, w,0 ))
print(tme.run_pass_pipeline(f, "tf-lower-to-mlprogram-and-hlo"))

# print('\n#3')
# x = tf.constant([[[0,1],[2,3]],[[4,5],[6,7]]], dtype=tf.float32); print(f"{x=}")
# w = tf.constant([[0,1],[1,0]], dtype=tf.int32); print(f"{w=}")
# y = tf.gather_nd(x, w, 0); print(f"{y=}")
# f =  tme.convert_function((tf.function(tf.gather_nd)).get_concrete_function(x, w, 0))
# print(tme.run_pass_pipeline(f, "tf-lower-to-mlprogram-and-hlo"))


# print('\n#4')
# x = tf.constant([[[0,1],[2,3]],[[4,5],[6,7]]], dtype=tf.float32); print(f"{x=}")
# w = tf.constant([[0,1],[1,0]], dtype=tf.int32); print(f"{w=}")
# y = tf.gather_nd(x, w, 0); print(f"{y=}")
# f =  tme.convert_function((tf.function(tf.gather_nd)).get_concrete_function(x, w, 0))
# print(tme.run_pass_pipeline(f, "tf-lower-to-mlprogram-and-hlo"))

# print('\n#5')
# x = tf.constant([[[0,1],[2,3]],[[4,5],[6,7]]], dtype=tf.float32); print(f"{x=}")
# w = tf.constant([[1],[0]], dtype=tf.int32); print(f"{w=}")
# y = tf.gather_nd(x, w, 1); print(f"{y=}")
# f =  tme.convert_function((tf.function(tf.gather_nd)).get_concrete_function(x, w, 1))
# print(tme.run_pass_pipeline(f, "tf-lower-to-mlprogram-and-hlo"))

# def translate_gather_nd_to_stablehlo(x, indices, axis):
#     input_rank = len(x.shape)
#     indices_rank = len(indices.shape)
    
#     # Determine the number of dimensions to collapse
#     collapsed_dims = list(range(indices_rank - 1))
    
#     # Determine the offset_dims and start_index_map based on the axis
#     offset_dims = [axis]
#     start_index_map = [axis]
    
#     # Translate indices for stablehlo.gather
#     translated_indices = indices
#     if indices_rank < input_rank:
#         translated_indices = tf.concat([tf.zeros_like(indices[:, :1]), indices], axis=1)
    
#     # Create the slice_sizes tensor for stablehlo.gather
#     slice_sizes = tf.concat([tf.ones_like(offset_dims), tf.shape(translated_indices)[1:]], axis=0)

#     params = {
#         'input_tensor': x,
#         'indices_tensor': translated_indices,
#         'dimension_numbers': {
#             'offset_dims': offset_dims,
#             'collapsed_slice_dims': collapsed_dims,
#             'start_index_map': start_index_map,
#             'index_vector_dim': indices_rank - 1,
#         },
#         'indices_are_sorted': False,
#         'slice_sizes': slice_sizes,
#     }

#     return params
# translated_result = translate_gather_nd_to_stablehlo(x, w, 0)
# print(translated_result)

# # Example #2
# x2 = tf.constant([[0., 1.], [2., 3.]], dtype=tf.float32)
# w2 = tf.constant([[1], [0]], dtype=tf.int32)
# axis2 = 0
# mlir_code2 = translate_gather_nd_to_stablehlo(x2, w2, axis2)
# print(mlir_code2)

# # Example #3
# x3 = tf.constant([[[0, 1], [2, 3]], [[4, 5], [6, 7]]], dtype=tf.float32)
# w3 = tf.constant([[0, 1], [1, 0]], dtype=tf.int32)
# axis3 = 0
# mlir_code3 = translate_gather_nd_to_stablehlo(x3, w3, axis3)
# print(mlir_code3)

# # Example #4
# x4 = tf.constant([[[0, 1], [2, 3]], [[4, 5], [6, 7]]], dtype=tf.float32)
# w4 = tf.constant([[0, 1], [1, 0]], dtype=tf.int32)
# axis4 = 0
# mlir_code4 = translate_gather_nd_to_stablehlo(x4, w4, axis4)
# print(mlir_code4)

# # Example #5
# x5 = tf.constant([[[0, 1], [2, 3]], [[4, 5], [6, 7]]], dtype=tf.float32)
# w5 = tf.constant([[1], [0]], dtype=tf.int32)
# axis5 = 1
# mlir_code5 = translate_gather_nd_to_stablehlo(x5, w5, axis5)
# print(mlir_code5)



import slope
from lib.models.cv.resnet_cifar import resnet

model = resnet(depth=20)
# model = resnet(depth=8)
N = 100
y = slope.arange(10, dtype=slope.float32)[None].expand((N // 10, 10)).reshape(-1)
x = slope.randn(N, 3, 32, 32)
optimizer = slope.nn.SGD(model, lr=0.1, momentum=0.9, weight_decay=1e-4)

def train_loss_fn(model, x, y):
    logits, model = model(x, training=True)
    # logits = model(x, training=False)
    loss = logits.cross_entropy(y) / x.size(0)
    return loss, ()#(model, logits)
    # return loss, (model, logits)

slope.jit(train_loss_fn)(model, x, y)
with slope.core.Timing(f"RUN: "): slope.jit(train_loss_fn)(model, x, y)
grad_fn = slope.jit(slope.value_and_grad(train_loss_fn, has_aux=True))
grad_fn(model, x, y)
with slope.core.Timing(f"GRAD: "): grad_fn(model, x, y)
with slope.core.Profiling(f"GRAD: "): grad_fn(model, x, y)

# with slope.core.Profiling(): loss, logits, model, optimizer, gmodel = train_step(model, (x, y), optimizer)


import os, gzip
import numpy as np
import struct
import array
from urllib import request

SLOPE_DATA = os.environ.get("SLOPE_DATA", os.path.join(os.path.expanduser("~"), "slope_data"))


def get_mnist():
    def download(url, save_path):
        if not os.path.exists(save_path_dir := os.path.dirname(save_path)):
            os.makedirs(save_path_dir, exist_ok=True)
        if not os.path.isfile(save_path):
            print(f"downloading {url}")
            request.urlretrieve(url, save_path)
            print(f"saved to {save_path}")

    def parse_labels(filename):
        with gzip.open(filename, "rb") as fh:
            _ = struct.unpack(">II", fh.read(8))
            return np.array(array.array("B", fh.read()), dtype=np.uint8)

    def parse_images(filename):
        with gzip.open(filename, "rb") as fh:
            _, num_data, rows, cols = struct.unpack(">IIII", fh.read(16))
            return np.array(array.array("B", fh.read()), dtype=np.uint8).reshape(num_data, rows, cols)

    base_url = "https://storage.googleapis.com/cvdf-datasets/mnist/"
    for filename in [
        "train-images-idx3-ubyte.gz",
        "train-labels-idx1-ubyte.gz",
        "t10k-images-idx3-ubyte.gz",
        "t10k-labels-idx1-ubyte.gz",
    ]:
        download(base_url + filename, os.path.join(SLOPE_DATA, filename))

    train_images = parse_images(os.path.join(SLOPE_DATA, "train-images-idx3-ubyte.gz"))
    train_labels = parse_labels(os.path.join(SLOPE_DATA, "train-labels-idx1-ubyte.gz"))
    test_images = parse_images(os.path.join(SLOPE_DATA, "t10k-images-idx3-ubyte.gz"))
    test_labels = parse_labels(os.path.join(SLOPE_DATA, "t10k-labels-idx1-ubyte.gz"))

    train_images = train_images / np.float32(255.0)
    train_labels = np.array(train_labels)
    test_images = test_images / np.float32(255.0)
    test_labels = np.array(test_labels)

    return train_images, train_labels, test_images, test_labels


if __name__ == "__main__":
    train_images, train_labels, test_images, test_labels = get_mnist()
    breakpoint()



import slope
from slope.core import Tensor

import slope.nn as nn

import time
import itertools
import math
import numpy as np
from tqdm import tqdm
from typing import Dict, Any

import numpy as np

BS, EVAL_BS, STEPS = 512, 50, 1000
np_dtype = np.float32
bias_scaler = 58
hyp: Dict[str, Any] = {
    "seed": 209,
    "opt": {
        "bias_lr": 1.76 * bias_scaler / 512,
        "non_bias_lr": 1.76 / 512,
        "bias_decay": 1.08 * 6.45e-4 * BS / bias_scaler,
        "non_bias_decay": 1.08 * 6.45e-4 * BS,
        "final_lr_ratio": 0.025,
        "initial_div_factor": 1e16,
        "label_smoothing": 0.20,
        "momentum": 0.85,
        "percent_start": 0.23,
        "loss_scale_scaler": 1.0 / 128,  # (range: ~1/512 - 16+, 1/128 w/ FP16)
    },
    "net": {
        "kernel_size": 2,  # kernel size for the whitening layer
        "cutmix_size": 3,
        "cutmix_steps": 499,
        "pad_amount": 2,
    },
    "ema": {
        "steps": 399,
        "decay_base": 0.95,
        "decay_pow": 1.6,
        "every_n_steps": 5,
    },
}


def whitening(X, kernel_size=hyp["net"]["kernel_size"]):
    def _cov(X):
        X = X / np.sqrt(X.shape[0] - 1)
        return X.T @ X

    def _patches(data, patch_size=(kernel_size, kernel_size)):
        h, w = patch_size
        c = data.shape[1]
        dim: SupportsIndex = (2, 3)  # type: ignore
        return (
            np.lib.stride_tricks.sliding_window_view(data, window_shape=(h, w), dim=dim)
            .transpose((0, 3, 2, 1, 4, 5))
            .reshape((-1, c, h, w))
        )

    def _eigens(patches):
        n, c, h, w = patches.shape
        Σ = _cov(patches.reshape(n, c * h * w))
        Λ, V = np.linalg.eigh(Σ, UPLO="U")
        return np.flip(Λ, 0), np.flip(V.T.reshape(c * h * w, c, h, w), 0)

    Λ, V = _eigens(_patches(X.numpy()))
    W = V / np.sqrt(Λ + 1e-2)[:, None, None, None]

    return Tensor(W.astype(np_dtype), requires_grad=False)


class ConvGroup:
    def __init__(self, channels_in, channels_out):
        self.conv1 = nn.Conv2d(channels_in, channels_out, kernel_size=3, padding=1, bias=False)
        self.conv2 = nn.Conv2d(channels_out, channels_out, kernel_size=3, padding=1, bias=False)

        self.norm1 = nn.BatchNorm(channels_out)
        self.norm2 = nn.BatchNorm(channels_out)

    def __call__(self, x):
        x = self.conv1(x)
        x = x.max_pool2d(2)
        x = x.float()
        x = self.norm1(x)
        x = x.gelu()
        residual = x
        x = self.conv2(x)
        x = x.float()
        x = self.norm2(x)
        x = x.gelu()
        return x + residual


class SpeedyResNet:
    def __init__(self, W):
        self.whitening = W
        self.net = [
            nn.Conv2d(12, 32, kernel_size=1, bias=False),
            lambda x: x.gelu(),
            ConvGroup(32, 64),
            ConvGroup(64, 256),
            ConvGroup(256, 512),
            lambda x: x.max((2, 3)),
            nn.Linear(512, 10, bias=False),
            lambda x: x.mul(1.0 / 9),
        ]

    def __call__(self, x, training=True):
        forward = lambda x: x.conv2d(self.whitening).pad2d((1, 0, 0, 1)).sequential(self.net)
        return forward(x) if training else forward(x) * 0.5 + forward(x[..., ::-1]) * 0.5


if __name__ == "__main__":
    net = SpeedyResNet(whitening)



import tensor  # python tensor

"""
A lightweight, pure Python, numpy compliant Tensor class.

The documenation in this module is rather compact. For details on each
function, see the corresponding documentation at:
http://docs.scipy.org/doc/numpy/reference/index.html Be aware that the
behavior of tinynumpy may deviate in some ways from numpy, or that
certain features may not be supported.
"""

# todo: keep track of readonly better
# todo: mathematical operators
# todo: more methods?
# todo: logspace, meshgrid
# todo: Fortran order?


import sys
import ctypes

from math import sqrt
from copy import copy, deepcopy
from typing import Iterable
import operator


# Python 2/3 compat
if sys.version_info >= (3,):
    xrange = range

# Define version numer
__version__ = "0.0.1dev"

# Define dtypes: struct name, short name, numpy name, ctypes type
_dtypes = [
    ("B", "b1", "bool", ctypes.c_bool),
    ("b", "i1", "int8", ctypes.c_int8),
    ("B", "u1", "uint8", ctypes.c_uint8),
    ("h", "i2", "int16", ctypes.c_int16),
    ("H", "u2", "uint16", ctypes.c_uint16),
    ("i", "i4", "int32", ctypes.c_int32),
    ("I", "u4", "uint32", ctypes.c_uint32),
    ("q", "i8", "int64", ctypes.c_int64),
    ("Q", "u8", "uint64", ctypes.c_uint64),
    ("f", "f4", "float32", ctypes.c_float),
    ("d", "f8", "float64", ctypes.c_double),
]

# Inject common dtype names
_known_dtypes = [d[2] for d in _dtypes]
for d in _known_dtypes:
    globals()[d] = d

newdim = None

nan = float("nan")


def _convert_dtype(dtype, to="numpy"):
    """Convert dtype, if could not find, pass as it was."""
    if dtype is None:
        return dtype
    dtype = str(dtype)
    index = {"Tensor": 0, "short": 1, "numpy": 2, "ctypes": 3}[to]
    for dd in _dtypes:
        if dtype in dd:
            return dd[index]
    return dtype  # Otherwise return original


def _ceildiv(a, b):
    return -(-a // b)


def _get_step(view):
    """Return step to walk over np. If 1, the Tensor is fully
    C-contiguous. If 0, the striding is such that one cannot
    step through the np.
    """
    cont_strides = _strides_for_shape(view.shape, view.itemsize)

    step = view.strides[-1] // cont_strides[-1]
    corrected_strides = tuple([i * step for i in cont_strides])

    almost_cont = view.strides == corrected_strides
    if almost_cont:
        return step
    else:
        return 0  # not contiguous


def _strides_for_shape(shape, itemsize):
    strides = []
    stride_product = 1
    for s in reversed(shape):
        strides.append(stride_product)
        stride_product *= s
    return tuple([i * itemsize for i in reversed(strides)])


def _size_for_shape(shape):
    stride_product = 1
    for s in shape:
        stride_product *= s
    return stride_product


def squeeze_strides(s):
    """Pop strides for singular dimensions."""
    return tuple([s[0]] + [s[i] for i in range(1, len(s)) if s[i] != s[i - 1]])


def _shape_from_object(obj):
    shape = []

    # todo: make more efficient, use len() etc
    def _shape_from_object_r(index, element, dim):
        try:
            for i, e in enumerate(element):
                _shape_from_object_r(i, e, dim + 1)
            while len(shape) <= dim:
                shape.append(0)
            l = i + 1
            s = shape[dim]
            if l > s:
                shape[dim] = l
        except TypeError:
            pass

    _shape_from_object_r(0, obj, 0)
    return tuple(shape)


def _assign_from_object(Tensor, obj):
    key = []

    # todo: make more efficient, especially the try-except
    def _assign_from_object_r(element):
        try:
            for i, e in enumerate(element):
                key.append(i)
                _assign_from_object_r(e)
                key.pop()
        except TypeError:
            Tensor[tuple(key)] = element

    _assign_from_object_r(obj)


def _increment_mutable_key(key, shape):
    for dim in reversed(xrange(len(shape))):
        key[dim] += 1
        if key[dim] < shape[dim]:
            return True
        if dim == 0:
            return False
        key[dim] = 0


def _key_for_index(index, shape):
    key = []
    cumshape = [1]
    for i in reversed(shape):
        cumshape.insert(0, cumshape[0] * i)
    for s in cumshape[1:-1]:
        n = index // s
        key.append(n)
        index -= n * s
    key.append(index)
    return tuple(key)


def _zerositer(n):
    for i in xrange(n):
        yield 0


## Public functions


def tensor(obj, dtype=None, copy=True, order=None):
    """Tensor(obj, dtype=None, copy=True, order=None)

    Create a new np. If obj is an Tensor, and copy=False, a view
    of that Tensor is returned. For details see:
    http://docs.scipy.org/doc/numpy/reference/generated/numpy.np.html
    """
    dtype = _convert_dtype(dtype)

    if isinstance(obj, Tensor):
        # From existing Tensor
        a = obj.view()
        if dtype is not None and dtype != a.dtype:
            a = a.astype(dtype)
        elif copy:
            a = a.copy()
        return a
    if hasattr(obj, "__Tensor_interface__"):
        # From something that looks like an Tensor, we can create
        # the ctypes Tensor for this and use that as a buffer
        D = obj.__Tensor_interface__
        # Get dtype
        dtype_orig = _convert_dtype(D["typestr"][1:])
        # Create Tensor
        if D["strides"]:
            itemsize = int(D["typestr"][-1])
            bufsize = D["strides"][0] * D["shape"][0] // itemsize
        else:
            bufsize = _size_for_shape(D["shape"])

        BufType = _convert_dtype(dtype_orig, "ctypes") * bufsize
        buffer = BufType.from_address(D["data"][0])
        a = Tensor(D["shape"], dtype_orig, buffer=buffer, strides=D["strides"], order=order)
        # Convert or copy?
        if dtype is not None and dtype != dtype_orig:
            a = a.astype(dtype)
        elif copy:
            a = a.copy()
        return a
    else:
        # From some kind of iterable
        shape = _shape_from_object(obj)
        # Try to derive dtype
        if dtype is None:
            el = obj
            while isinstance(el, (tuple, list)) and el:
                el = el[0]
            if isinstance(el, int):
                dtype = "int64"
        # Create Tensor
        a = Tensor(shape, dtype, order=None)
        _assign_from_object(a, obj)
        return a


def zeros_like(a, dtype=None, order=None):
    """Return an Tensor of zeros with the same shape and type as a given np."""
    dtype = a.dtype if dtype is None else dtype
    return zeros(a.shape, dtype, order)


def ones_like(a, dtype=None, order=None):
    """Return an Tensor of ones with the same shape and type as a given np."""
    dtype = a.dtype if dtype is None else dtype
    return ones(a.shape, dtype, order)


def empty_like(a, dtype=None, order=None):
    """Return a new Tensor with the same shape and type as a given np."""
    dtype = a.dtype if dtype is None else dtype
    return empty(a.shape, dtype, order)


def zeros(shape, dtype=None, order=None):
    """Return a new Tensor of given shape and type, filled with zeros"""
    return empty(shape, dtype, order)


def ones(shape, dtype=None, order=None):
    """Return a new Tensor of given shape and type, filled with ones"""
    a = empty(shape, dtype, order)
    a.fill(1)
    return a


def eye(size):
    """Return a new 2d Tensor with given dimensions, filled with ones on the
    diagonal and zeros elsewhere.
    """
    a = zeros((size, size))
    for i in xrange(size):
        a[i, i] = 1
    return a


def empty(shape, dtype=None, order=None):
    """Return a new Tensor of given shape and type, without initializing entries"""
    return Tensor(shape, dtype, order=order)


def arange(*args, **kwargs):
    """arange([start,] stop[, step,], dtype=None)

    Return evenly spaced values within a given interval.

    Values are generated within the half-open interval ``[start, stop)``
    (in other words, the interval including `start` but excluding `stop`).
    For integer arguments the function is equivalent to the Python built-in
    `range <http://docs.python.org/lib/built-in-funcs.html>`_ function,
    but returns an Tensor rather than a list.

    When using a non-integer step, such as 0.1, the results will often not
    be consistent.  It is better to use ``linspace`` for these cases.
    """
    # Get dtype
    dtype = kwargs.pop("dtype", None)
    if kwargs:
        x = list(kwargs.keys())[0]
        raise TypeError("arange() got an unexpected keyword argument %r" % x)
    # Parse start, stop, step
    if len(args) == 0:
        raise TypeError('Required argument "start" not found')
    elif len(args) == 1:
        start, stop, step = 0, int(args[0]), 1
    elif len(args) == 2:
        start, stop, step = int(args[0]), int(args[1]), 1
    elif len(args) == 3:
        start, stop, step = int(args[0]), int(args[1]), int(args[2])
    else:
        raise TypeError("Too many input arguments")
    # Init
    iter = xrange(start, stop, step)
    a = empty((len(iter),), dtype=dtype)
    a[:] = list(iter)
    return a


def linspace(start, stop, num=50, endpoint=True, retstep=False, dtype=None):
    """linspace(start, stop, num=50, endpoint=True, retstep=False, dtype=None)

    Return evenly spaced numbers over a specified interval. Returns num
    evenly spaced samples, calculated over the interval [start, stop].
    The endpoint of the interval can optionally be excluded.
    """
    # Prepare
    start, stop = float(start), float(stop)
    ra = stop - start
    if endpoint:
        step = ra / (num - 1)
    else:
        step = ra / num
    # Create
    a = empty((num,), dtype)
    a[:] = [start + i * step for i in xrange(num)]
    # Return
    if retstep:
        return a, step
    else:
        return a


def add(Tensor_vec1, Tensor_vec2):
    c = []
    for a, b in zip(Tensor_vec1, Tensor_vec2):
        c.append(a + b)
    cRay = Tensor(c)
    return cRay


def subtract(Tensor_vec1, Tensor_vec2):
    c = []
    for a, b in zip(Tensor_vec1, Tensor_vec2):
        c.append(a - b)
    cRay = Tensor(c)
    return cRay


def multiply(Tensor_vec1, Tensor_vec2):
    c = []
    for a, b in zip(Tensor_vec1, Tensor_vec2):
        c.append(a * b)
    cRay = Tensor(c)
    return cRay


def divide(Tensor_vec1, integer):
    c = []
    for a in Tensor_vec1:
        c.append(a / integer)
    cRay = Tensor(c)
    return cRay


class LinAlgError(Exception):
    pass


def det(A):
    if len(A) == 3 and [len(vec) == 3 for vec in A]:
        try:
            # http://mathworld.wolfram.com/Determinant.html
            det_A = (
                A[0][0] * A[1][1] * A[2][2]
                + A[0][1] * A[1][2] * A[2][0]
                + A[0][2] * A[1][0] * A[2][1]
                - (A[0][2] * A[1][1] * A[2][0] + A[0][1] * A[1][0] * A[2][2] + A[0][0] * A[1][2] * A[2][1])
            )
        except LinAlgError as e:
            det_A = e
    else:
        raise IndexError("Vector has invalid dimensions")
    return det_A


def cross(u, v):
    """
    Return the cross product of two 2 or 3 dimensional vectors.
    """

    uDim = len(u)
    vDim = len(v)

    uxv = []

    # http://mathworld.wolfram.com/CrossProduct.html
    if uDim == vDim == 2:
        try:
            uxv = [u[0] * v[1] - u[1] * v[0]]
        except LinAlgError as e:
            uxv = e
    elif uDim == vDim == 3:
        try:
            for i in range(uDim):
                uxv = [
                    u[1] * v[2] - u[2] * v[1],
                    -(u[0] * v[2] - u[2] * v[0]),
                    u[0] * v[1] - u[1] * v[0],
                ]
        except LinAlgError as e:
            uxv = e
    else:
        raise IndexError("Vector has invalid dimensions")
    return uxv


def dot(u, v):
    """
    Return the dot product of two equal-dimensional vectors.
    """

    uDim = len(u)
    vDim = len(v)

    # http://reference.wolfram.com/language/ref/Dot.html
    if uDim == vDim:
        try:
            u_dot_v = sum(map(operator.mul, u, v))
        except LinAlgError as e:
            u_dot_v = e
    else:
        raise IndexError("Vector has invalid dimensions")
    return u_dot_v


def reshape(X, shape):
    """
    Returns the reshaped image of an Tensor
    """
    assert isinstance(X, Tensor)
    assert isinstance(shape, tuple) or isinstance(shape, list)
    return X.reshape(shape)


## The class


class Tensor(object):
    """Tensor(shape, dtype='float64', buffer=None, offset=0,
                strides=None, order=None)

    Tensor class similar to numpy's Tensor, implemented in pure Python.
    This class can be distinguished from a real numpy Tensor in that
    the repr always shows the dtype as a string, and for larger Tensors
    (more than 100 elements) it shows a short one-line repr.

    An Tensor object represents a multidimensional, homogeneous Tensor
    of fixed-size items.  An associated data-type property describes the
    format of each element in the np.

    Tensors should be constructed using `Tensor`, `zeros` or `empty` (refer
    to the See Also section below).  The parameters given here refer to
    a low-level method (`Tensor(...)`) for instantiating an np.

    Parameters
    ----------
    shape : tuple of ints
        Shape of created np.
    dtype : data-type, optional
        Any object that can be interpreted as a numpy data type.
    buffer : object contaning data, optional
        Used to fill the Tensor with data. If another Tensor is given,
        the underlying data is used. Can also be a ctypes.Tensor or any
        object that exposes the buffer interface.
    offset : int, optional
        Offset of Tensor data in buffer.
    strides : tuple of ints, optional
        Strides of data in memory.
    order : {'C', 'F'}, optional  NOT SUPPORTED
        Row-major or column-major order.

    Attributes
    ----------
    T : Tensor
        permute of the np. In tinynumpy only supported for ndim <= 3.
    data : buffer
        The Tensor's elements, in memory. In tinynumpy this is a ctypes np.
    dtype : str
        Describes the format of the elements in the np. In tinynumpy
        this is a string.
    flags : dict
        Dictionary containing information related to memory use, e.g.,
        'C_CONTIGUOUS', 'OWNDATA', 'WRITEABLE', etc.
    flat : iterator object
        Flattened version of the Tensor as an iterator. In tinynumpy
        the iterator cannot be indexed.
    size : int
        Number of elements in the np.
    itemsize : int
        The memory use of each Tensor element in bytes.
    nbytes : int
        The total number of bytes required to store the Tensor data,
        i.e., ``itemsize * size``.
    ndim : int
        The Tensor's number of dimensions.
    shape : tuple of ints
        Shape of the np.
    strides : tuple of ints
        The step-size required to move from one element to the next in
        memory. For example, a contiguous ``(3, 4)`` Tensor of type
        ``int16`` in C-order has strides ``(8, 2)``.  This implies that
        to move from element to element in memory requires jumps of 2 bytes.
        To move from row-to-row, one needs to jump 8 bytes at a time
        (``2 * 4``).
    base : Tensor
        If the Tensor is a view into another Tensor, that Tensor is its `base`
        (unless that Tensor is also a view).  The `base` Tensor is where the
        Tensor data is actually stored.
    __Tensor_interface__ : dict
        Dictionary with low level Tensor information. Used by numpy to
        turn into a real numpy np. Can also be used to give C libraries
        access to the data via ctypes.

    See Also
    --------
    Tensor : Construct an np.
    zeros : Create an Tensor, each element of which is zero.
    empty : Create an Tensor, but leave its allocated memory unchanged (i.e.,
            it contains "garbage").

    Notes
    -----
    There are two modes of creating an Tensor:

    1. If `buffer` is None, then only `shape`, `dtype`, and `order`
       are used.
    2. If `buffer` is an object exposing the buffer interface, then
       all keywords are interpreted.

    """

    __slots__ = [
        "_dtype",
        "_shape",
        "_strides",
        "_itemsize",
        "_offset",
        "_base",
        "_data",
    ]

    def __init__(self, shape, dtype="float64", buffer=None, offset=0, strides=None, order=None):
        # Check order
        if order is not None:
            raise MachineError("Tensor order parameter is not supported")
        # Check and set shape
        try:
            assert isinstance(shape, Iterable)
            shape = tuple(shape)
        except Exception as e:
            raise AssertionError("The shape must be tuple or list")
        assert all([isinstance(x, int) for x in shape])
        self._shape = shape
        # Check and set dtype
        dtype = _convert_dtype(dtype) if (dtype is not None) else "float64"
        if dtype not in _known_dtypes:
            raise TypeError("data type %r not understood" % dtype)
        self._dtype = dtype
        # Itemsize is directly derived from dtype
        self._itemsize = int(_convert_dtype(dtype, "short")[-1])

        if buffer is None:
            # New Tensor
            self._base = None
            # Check and set offset and strides
            assert offset == 0
            self._offset = 0
            assert strides is None
            self._strides = _strides_for_shape(self._shape, self.itemsize)

        else:
            # Existing Tensor
            if isinstance(buffer, Tensor) and buffer.base is not None:
                buffer = buffer.base
            # Keep a reference to avoid memory cleanup
            self._base = buffer
            # for Tensor we use the data property
            if isinstance(buffer, Tensor):
                buffer = buffer.data
            # Check and set offset
            assert isinstance(offset, int) and offset >= 0
            self._offset = offset
            # Check and set strides
            if strides is None:
                strides = _strides_for_shape(shape, self.itemsize)
            assert isinstance(strides, tuple)
            assert all([isinstance(x, int) for x in strides])
            assert len(strides) == len(shape)
            self._strides = strides

        # Define our buffer class
        buffersize = self._strides[0] * self._shape[0] // self._itemsize
        buffersize += self._offset
        BufferClass = _convert_dtype(dtype, "ctypes") * buffersize
        # Create buffer
        if buffer is None:
            self._data = BufferClass()
        elif isinstance(buffer, ctypes.Tensor):
            self._data = BufferClass.from_address(ctypes.addressof(buffer))
        else:
            self._data = BufferClass.from_buffer(buffer)

    @property
    def __Tensor_interface__(self):
        """Allow converting to real numpy Tensor, or pass pointer to C library
        http://docs.scipy.org/doc/numpy/reference/Tensors.interface.html
        """
        readonly = False
        # typestr
        typestr = "<" + _convert_dtype(self.dtype, "short")
        # Pointer
        if isinstance(self._data, ctypes.Tensor):
            ptr = ctypes.addressof(self._data)
        elif hasattr(self._data, "__Tensor_interface__"):
            ptr, readonly = self._data.__Tensor_interface__["data"]
        elif hasattr(self._data, "buffer_info"):  # Python's np.Tensor
            ptr = self._data.buffer_info()[0]
        elif isinstance(self._data, bytes):
            ptr = ctypes.cast(self._data, ctypes.c_symp).value
            readonly = True
        else:
            raise TypeError("Cannot get address to underlying Tensor data")
        ptr += self._offset * self.itemsize
        #
        return dict(
            version=3,
            shape=self.shape,
            typestr=typestr,
            descr=[("", typestr)],
            data=(ptr, readonly),
            strides=self.strides,
            # offset=self._offset,
            # mask=None,
        )

    def __len__(self):
        return self.shape[0]

    def __getitem__(self, key):
        offset, shape, strides = self._index_helper(key)
        if not shape:
            # Return scalar
            return self._data[offset]
        else:
            # Return view
            return Tensor(shape, self.dtype, offset=offset, strides=strides, buffer=self)

    def __setitem__(self, key, value):
        # Get info for view
        offset, shape, strides = self._index_helper(key)

        # Is this easy?
        if not shape:
            self._data[offset] = value
            return

        # Create view to set data to
        view = Tensor(shape, self.dtype, offset=offset, strides=strides, buffer=self)

        # Get data to set as a list (because getting slices from ctype
        # Tensors yield lists anyway). The list is our "contiguous Tensor"
        if isinstance(value, (float, int)):
            value_list = [value] * view.size
        elif isinstance(value, (tuple, list)):
            value_list = value
        else:
            if not isinstance(value, Tensor):
                value = Tensor(value, copy=False)
            value_list = value._toflatlist()

        # Check if size match
        if view.size != len(value_list):
            raise ValueError("Number of elements in source does not match " "number of elements in target.")

        # Assign data in most efficient way that we can. This code
        # looks for the largest semi-contiguous block: the block that
        # we can access as a 1D Tensor with a stepsize.
        subviews = [view]
        value_index = 0
        count = 0
        while subviews:
            subview = subviews.pop(0)
            step = _get_step(subview)
            if step:
                block = value_list[value_index : value_index + subview.size]
                s = slice(subview._offset, subview._offset + subview.size * step, step)
                view._data[s] = block
                value_index += subview.size
                count += 1
            else:
                for i in range(subview.shape[0]):
                    subviews.append(subview[i])
        assert value_index == len(value_list)

    def __float__(self):
        if self.size == 1:
            return float(self.data[self._offset])
        else:
            raise TypeError("Only length-1 Tensors can be converted to scalar")

    def __int__(self):
        if self.size == 1:
            return int(self.data[self._offset])
        else:
            raise TypeError("Only length-1 Tensors can be converted to scalar")

    def __repr__(self):
        # If more than 100 elements, show short repr
        if self.size > 100:
            shapestr = "x".join([str(i) for i in self.shape])
            return "<Tensor %s %s at 0x%x>" % (shapestr, self.dtype, id(self))

        # Otherwise, try to show in nice way
        def _repr_r(s, dim, offset):
            dimindent = min(2, max(0, (self.ndim - dim - 1)))
            if dim < len(self.shape):
                s += "["
                for k_index, k in enumerate(xrange(self.shape[dim])):
                    if k_index > 0:
                        s += ("\n       " + " " * dim) * dimindent
                    offset_ = offset + k * self._strides[dim] // self.itemsize
                    s = _repr_r(s, dim + 1, offset_)
                    if k_index < self.shape[dim] - 1:
                        s += ", "
                s += "]"
            else:
                r = repr(self.data[offset])
                if "." in r:
                    r = " " + r
                    if r.endswith(".0"):
                        r = r[:-1]
                s += r
            return s

        s = _repr_r("", 0, self._offset)
        if self.dtype != "float64" and self.dtype != "int32":
            return "Tensor(" + s + ", dtype='%s')" % self.dtype
        else:
            return "Tensor(" + s + ")"

    def __eq__(self, other):
        if other.__module__.split(".")[0] == "numpy":
            return other == self
        else:
            out = empty(self.shape, "bool")
            out[:] = [i1 == i2 for (i1, i2) in zip(self.flat, other.flat)]
            return out

    def __add__(self, other):
        """classic addition"""
        if isinstance(other, int) or isinstance(other, float):
            out = empty(self.shape, self.dtype)
            out[:] = [dat + other for dat in self._data]
            return out
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                out = empty(self.shape, self.dtype)
                out[:] = [i + j for (i, j) in zip(self.flat, other.flat)]
                return out

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            out = empty(self.shape, self.dtype)
            out[:] = [dat - other for dat in self._data]
            return out
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                out = empty(self.shape, self.dtype)
                out[:] = [i - j for (i, j) in zip(self.flat, other.flat)]
                return out

    def __rsub__(self, other):
        return self.__sub__(other)

    def __mul__(self, other):
        """multiply element-wise with Tensor or float/scalar"""
        if isinstance(other, int) or isinstance(other, float):
            out = empty(self.shape, self.dtype)
            out[:] = [dat * other for dat in self._data]
            return out
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                out = empty(self.shape, self.dtype)
                out[:] = [i * j for (i, j) in zip(self.flat, other.flat)]
                return out

    def __rmul__(self, other):
        return self.__mul__(other)

    def __div__(self, other):
        """divide element-wise with Tensor or float/scalar"""
        if isinstance(other, int) or isinstance(other, float):
            if other == 0:
                raise ZeroDivisionError
            out = empty(self.shape, self.dtype)
            out[:] = [dat / other for dat in self._data]
            return out
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                out = empty(self.shape, self.dtype)
                out[:] = [i / j for (i, j) in zip(self.flat, other.flat)]
                return out
            else:
                raise ValueError("Tensor sizes do not match. " + str(self.shape) + " versus " + str(other.shape))

    def __truediv__(self, other):
        """divide element-wise with Tensor or float/scalar"""
        if isinstance(other, int) or isinstance(other, float):
            if other == 0:
                raise ZeroDivisionError
            out = empty(self.shape, self.dtype)
            out[:] = [dat / other for dat in self._data]
            return out
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                out = empty(self.shape, self.dtype)
                out[:] = [i / j for (i, j) in zip(self.flat, other.flat)]
                return out
            else:
                raise ValueError("Tensor sizes do not match. " + str(self.shape) + " versus " + str(other.shape))

    def __floordiv__(self, other):
        """divide element-wise with Tensor or float/scalar"""
        if isinstance(other, int) or isinstance(other, float):
            if other == 0:
                raise ZeroDivisionError
            out = empty(self.shape, self.dtype)
            out[:] = [dat // other for dat in self._data]
            return out
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                out = empty(self.shape, self.dtype)
                out[:] = [i // j for (i, j) in zip(self.flat, other.flat)]
                return out
            else:
                raise ValueError("Tensor sizes do not match. " + str(self.shape) + " versus " + str(other.shape))

    def __mod__(self, other):
        """divide element-wise with Tensor or float/scalar"""
        if isinstance(other, int) or isinstance(other, float):
            out = empty(self.shape, self.dtype)
            out[:] = [dat % other for dat in self._data]
            return out
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                out = empty(self.shape, self.dtype)
                out[:] = [i % j for (i, j) in zip(self.flat, other.flat)]
                return out
            else:
                raise ValueError("Tensor sizes do not match. " + str(self.shape) + " versus " + str(other.shape))

    def __pow__(self, other):
        """power of two Tensors element-wise (of just float power)"""
        if isinstance(other, int) or isinstance(other, float):
            out = empty(self.shape, self.dtype)
            out[:] = [dat**other for dat in self._data]
            return out
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                out = empty(self.shape, self.dtype)
                out[:] = [i**j for (i, j) in zip(self.flat, other.flat)]
                return out
            else:
                raise ValueError("Tensor sizes do not match. " + str(self.shape) + " versus " + str(other.shape))

    def __iadd__(self, other):
        """Addition of other Tensor or float in place with += operator"""
        if isinstance(other, int) or isinstance(other, float):
            for i in range(len(self._data)):
                self._data[i] += other
            return self
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                for i in range(len(self._data)):
                    self._data[i] += other._data[i]
                return self
            else:
                raise ValueError("Tensor sizes do not match. " + str(self.shape) + " versus " + str(other.shape))

    def __isub__(self, other):
        """Addition of other Tensor or float in place with += operator"""
        if isinstance(other, int) or isinstance(other, float):
            for i in range(len(self._data)):
                self._data[i] -= other
            return self
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                for i in range(len(self._data)):
                    self._data[i] -= other._data[i]
                return self
            else:
                raise ValueError("Tensor sizes do not match. " + str(self.shape) + " versus " + str(other.shape))

    def __imul__(self, other):
        """multiplication woth other Tensor or float in place with *= operator"""
        if isinstance(other, int) or isinstance(other, float):
            for i in range(len(self._data)):
                self._data[i] *= other
            return self
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                for i in range(len(self._data)):
                    self._data[i] *= other._data[i]
                return self
            else:
                raise ValueError("Tensor sizes do not match. " + str(self.shape) + " versus " + str(other.shape))

    def __idiv__(self, other):
        """Division of other Tensor or float in place with /= operator"""
        if isinstance(other, int) or isinstance(other, float):
            if other == 0:
                raise ZeroDivisionError
            for i in range(len(self._data)):
                self._data[i] /= other
            return self
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                for i in range(len(self._data)):
                    self._data[i] /= other._data[i]
                return self
            else:
                raise ValueError("Tensor sizes do not match. " + str(self.shape) + " versus " + str(other.shape))

    def __itruediv__(self, other):
        """Division of other Tensor or float in place with /= operator"""
        if isinstance(other, int) or isinstance(other, float):
            if other == 0:
                raise ZeroDivisionError
            for i in range(len(self._data)):
                self._data[i] /= other
            return self
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                for i in range(len(self._data)):
                    self._data[i] /= other._data[i]
                return self
            else:
                raise ValueError("Tensor sizes do not match. " + str(self.shape) + " versus " + str(other.shape))

    def __ifloordiv__(self, other):
        """Division of other Tensor or float in place with /= operator"""
        if isinstance(other, int) or isinstance(other, float):
            if other == 0:
                raise ZeroDivisionError
            for i in range(len(self._data)):
                self._data[i] //= other
            return self
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                for i in range(len(self._data)):
                    self._data[i] //= other._data[i]
                return self
            else:
                raise ValueError("Tensor sizes do not match. " + str(self.shape) + " versus " + str(other.shape))

    def __imod__(self, other):
        """mod of other Tensor or float in place with /= operator"""
        if isinstance(other, int) or isinstance(other, float):
            if other == 0:
                raise ZeroDivisionError
            for i in range(len(self._data)):
                self._data[i] %= other
            return self
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                for i in range(len(self._data)):
                    self._data[i] %= other._data[i]
                return self
            else:
                raise ValueError("Tensor sizes do not match. " + str(self.shape) + " versus " + str(other.shape))

    def __imod__(self, other):
        """mod of other Tensor or float in place with /= operator"""
        if isinstance(other, int) or isinstance(other, float):
            if other == 0:
                raise ZeroDivisionError
            for i in range(len(self._data)):
                self._data[i] %= other
            return self
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                for i in range(len(self._data)):
                    self._data[i] %= other._data[i]
                return self
            else:
                raise ValueError("Tensor sizes do not match. " + str(self.shape) + " versus " + str(other.shape))

    def __ipow__(self, other):
        """mod of other Tensor or float in place with /= operator"""
        if isinstance(other, int) or isinstance(other, float):
            for i in range(len(self._data)):
                self._data[i] **= other
            return self
        if isinstance(other, Tensor):
            if self.shape == other.shape:
                for i in range(len(self._data)):
                    self._data[i] **= other._data[i]
                return self
            else:
                raise ValueError("Tensor sizes do not match. " + str(self.shape) + " versus " + str(other.shape))

    ## Private helper functions

    def _index_helper(self, key):
        # Indexing spec is located at:
        # http://docs.scipy.org/doc/numpy/reference/Tensors.indexing.html

        # Promote to tuple.
        if not isinstance(key, tuple):
            key = (key,)

        dim = 0
        shape = []
        strides = []
        offset = self._offset

        for k in key:
            dimsize = self._shape[dim]
            if isinstance(k, int):
                if k >= dimsize:
                    raise IndexError("index %i is out of bounds for dim %i " "with size %s" % (k, dim, dimsize))
                offset += k * self._strides[dim] // self.itemsize
                dim += 1
            elif isinstance(k, slice):
                start, stop, step = k.indices(self.shape[dim])
                shape.append(_ceildiv(stop - start, step))
                strides.append(step * self._strides[dim])
                offset += start * self._strides[dim] // self.itemsize
                dim += 1
            elif k is Ellipsis:
                raise TypeError("ellipsis are not supported.")
            elif k is None:
                shape.append(1)
                stride = 1
                for s in self._strides[dim:]:
                    stride *= s
                strides.append(stride)
            else:
                raise TypeError("key elements must be instaces of int or slice.")

        shape.extend(self.shape[dim:])
        strides.extend(self._strides[dim:])

        return offset, tuple(shape), tuple(strides)

    def _toflatlist(self):
        value_list = []
        subviews = [self]
        count = 0
        while subviews:
            subview = subviews.pop(0)
            step = _get_step(subview)
            if step:
                s = slice(subview._offset, subview._offset + subview.size * step, step)
                value_list += self._data[s]
                count += 1
            else:
                for i in range(subview.shape[0]):
                    subviews.append(subview[i])
        return value_list

    ## Properties

    @property
    def ndim(self):
        return len(self._shape)

    @property
    def size(self):
        return _size_for_shape(self._shape)

    @property
    def nbytes(self):
        return _size_for_shape(self._shape) * self.itemsize

    def _get_shape(self):
        return self._shape

    def _set_shape(self, newshape):
        if newshape == self.shape:
            return
        if self.size != _size_for_shape(newshape):
            raise ValueError("Total size of new Tensor must be unchanged")
        if _get_step(self) == 1:
            # Contiguous, hooray!
            self._shape = tuple(newshape)
            self._strides = _strides_for_shape(self._shape, self.itemsize)
            return

        # Else, try harder ... This code supports adding /removing
        # singleton dimensions. Although it may sometimes be possible
        # to split a dimension in two if the contiguous blocks allow
        # this, we don't bother with such complex cases for now.
        # Squeeze shape / strides
        N = self.ndim
        shape = [self.shape[i] for i in range(N) if self.shape[i] > 1]
        strides = [self.strides[i] for i in range(N) if self.shape[i] > 1]
        # Check if squeezed shapes match
        newshape_ = [newshape[i] for i in range(len(newshape)) if newshape[i] > 1]
        if newshape_ != shape:
            raise AttributeError("incompatible shape for non-contiguous Tensor")
        # Modify to make this data work in loop
        strides.append(strides[-1])
        shape.append(1)
        # Form new strides
        i = -1
        newstrides = []
        try:
            for s in reversed(newshape):
                if s == 1:
                    newstrides.append(strides[i] * shape[i])
                else:
                    i -= 1
                    newstrides.append(strides[i])
        except IndexError:
            # Fail
            raise AttributeError("incompatible shape for non-contiguous Tensor")
        else:
            # Success
            newstrides.reverse()
            self._shape = tuple(newshape)
            self._strides = tuple(newstrides)

    shape = property(_get_shape, _set_shape)  # Python 2.5 compat (e.g. Jython)

    @property
    def strides(self):
        return self._strides

    @property
    def dtype(self):
        return self._dtype

    @property
    def itemsize(self):
        return self._itemsize

    @property
    def base(self):
        return self._base

    @property
    def data(self):
        return self._data

    @property
    def flat(self):
        subviews = [self]
        count = 0
        while subviews:
            subview = subviews.pop(0)
            step = _get_step(subview)
            if step:
                s = slice(subview._offset, subview._offset + subview.size * step, step)
                for i in self._data[s]:
                    yield i
            else:
                for i in range(subview.shape[0]):
                    subviews.append(subview[i])

    @property
    def T(self):
        if self.ndim < 2:
            return self
        else:
            return self.permute()

    @property
    def flags(self):
        c_cont = _get_step(self) == 1
        return dict(
            C_CONTIGUOUS=c_cont,
            F_CONTIGUOUS=(c_cont and self.ndim < 2),
            OWNDATA=(self._base is None),
            WRITEABLE=True,  # todo: fix this
            ALIGNED=c_cont,  # todo: different from contiguous?
            UPDATEIFCOPY=False,  # We don't support this feature
        )

    ## Methods - managemenet

    def fill(self, value):
        assert isinstance(value, (int, float))
        self[:] = value

    def clip(self, a_min, a_max, out=None):
        if out is None:
            out = empty(self.shape, self.dtype)
        L = self._toflatlist()
        L = [min(a_max, max(a_min, x)) for x in L]
        out[:] = L
        return out

    def copy(self):
        out = empty(self.shape, self.dtype)
        out[:] = self
        return out

    def flatten(self):
        out = empty((self.size,), self.dtype)
        out[:] = self
        return out

    def ravel(self):
        return self.reshape((self.size,))

    def repeat(self, repeats, dim=None):
        if dim:
            raise (TypeError, "dim argument is not supported")
        out = empty((self.size * repeats,), self.dtype)
        for i in range(repeats):
            out[i * self.size : (i + 1) * self.size] = self
        return out

    def reshape(self, newshape):
        out = self.view()
        try:
            out.shape = newshape
        except AttributeError:
            out = self.copy()
            out.shape = newshape
        return out

    def permute(self):
        # Numpy returns a view, but we cannot do that since we do not
        # support Fortran ordering
        ndim = self.ndim
        if ndim < 2:
            return self.view()
        shape = self.shape[::-1]
        out = empty(shape, self.dtype)
        #
        if ndim == 2:
            for i in xrange(self.shape[0]):
                out[:, i] = self[i, :]
        elif ndim == 3:
            for i in xrange(self.shape[0]):
                for j in xrange(self.shape[1]):
                    out[:, j, i] = self[i, j, :]
        else:
            raise ValueError("Tinynumpy supports permute up to ndim=3")
        return out

    def astype(self, dtype):
        out = empty(self.shape, dtype, device)
        out[:] = self

    def view(self, dtype=None, type=None):
        if dtype is None:
            dtype = self.dtype
        if dtype == self.dtype:
            return Tensor(
                self.shape,
                dtype,
                buffer=self,
                offset=self._offset,
                strides=self.strides,
            )
        elif self.ndim == 1:
            itemsize = int(_convert_dtype(dtype, "short")[-1])
            size = self.nbytes // itemsize
            offsetinbytes = self._offset * self.itemsize
            offset = offsetinbytes // itemsize
            return Tensor((size,), dtype, buffer=self, offset=offset)
        else:
            raise ValueError("new type not compatible with np.")

    ## Methods - statistics

    # We use the self.flat generator here. self._toflatlist() would be
    # faster, but it might take up significantly more memory.

    def all(self, dim=None):
        if dim:
            raise (TypeError, "dim argument is not supported")
        return all(self.flat)

    def any(self, dim=None):
        if dim:
            raise (TypeError, "dim argument is not supported")
        return any(self.flat)

    def min(self, dim=None):
        if dim:
            raise (TypeError, "dim argument is not supported")
        return min(self.flat)

    def max(self, dim=None):
        if dim:
            raise (TypeError, "dim argument is not supported")
        return max(self.flat)
        # return max(self._toflatlist())  # almost twice as fast

    def sum(self, dim=None):
        if dim:
            raise (TypeError, "dim argument is not supported")
        return sum(self.flat)

    def prod(self, dim=None):
        if dim:
            raise (TypeError, "dim argument is not supported")
        p = 1.0
        for i in self.flat:
            p *= float(i)
        return p

    def ptp(self, dim=None):
        if dim:
            raise (TypeError, "dim argument is not supported")
        mn = self.data[self._offset]
        mx = mn
        for i in self.flat:
            if i > mx:
                mx = i
            if i < mn:
                mn = i
        return mx - mn

    def mean(self, dim=None):
        if dim:
            raise (TypeError, "dim argument is not supported")
        return self.sum() / self.size

    def argmax(self, dim=None):
        if dim:
            raise (TypeError, "dim argument is not supported")
        r = self.data[self._offset]
        r_index = 0
        for i_index, i in enumerate(self.flat):
            if i > r:
                r = i
                r_index = i_index
        return r_index

    def argmin(self, dim=None):
        if dim:
            raise (TypeError, "dim argument is not supported")
        r = self.data[self._offset]
        r_index = 0
        for i_index, i in enumerate(self.flat):
            if i < r:
                r = i
                r_index = i_index
        return r_index

    def cumprod(self, dim=None, out=None):
        if dim:
            raise (TypeError, "dim argument is not supported")
        if out is None:
            out = empty((self.size,), self.dtype)
        p = 1
        L = []
        for x in self.flat:
            p *= x
            L.append(p)
        out[:] = L
        return out

    def cumsum(self, dim=None, out=None):
        if dim:
            raise (TypeError, "dim argument is not supported")
        if out is None:
            out = empty((self.size,), self.dtype)
        p = 0
        L = []
        for x in self.flat:
            p += x
            L.append(p)
        out[:] = L
        return out

    def var(self, dim=None):
        if dim:
            raise (TypeError, "dim argument is not supported")
        m = self.mean()
        acc = 0
        for x in self.flat:
            acc += abs(x - m) ** 2
        return acc / self.size

    def std(self, dim=None):
        return sqrt(self.var(dim))

    def argwhere(self, val):
        # assumes that list has only values of same dtype

        idx = [i for i, e in enumerate(self.flat) if e == val]
        keys = [list(_key_for_index(i, self.shape)) for i in idx]
        return keys

    def tolist(self):
        """
        Returns the Tensor as a comprehensive list
        """
        shp = list(self.shape).copy()
        jump = self.size // shp[-1]
        n_comp = 0  # comprehension depth
        comp = list(self._data).copy()
        while n_comp < len(self.shape) - 1:
            comp = [comp[i * shp[-1] : i * shp[-1] + shp[-1]] for i in range(jump)]
            shp.pop()
            jump = len(comp) // shp[-1]
            n_comp += 1
        return comp


class nditer:
    def __init__(self, Tensor):
        self.Tensor = Tensor
        self.key = [0] * len(self.np.shape)

    def __iter__(self):
        return self

    def __len__(self):
        return _size_for_shape(self.np.shape)

    def __getitem__(self, index):
        key = _key_for_index(index, self.np.shape)
        return self.Tensor[key]

    def __next__(self):
        if self.key is None:
            raise StopIteration
        value = self.Tensor[tuple(self.key)]
        if not _increment_mutable_key(self.key, self.np.shape):
            self.key = None
        return value

    def next(self):
        return self.__next__()



import slope

from slope import jit, jvp, grad


def f(x):
    z = x * 2.0
    y = g(z)
    return y


def g(x):
    z = x.cos()
    y = z * 2.0
    return y


x = slope.tensor(3.0)
x_dot = slope.tensor(1.0)

y1 = f(x)
y2 = jit(f)(x)
y3, _ = jvp(f, (x,), (x_dot,))
y4, _ = jvp(jit(f), (x,), (x_dot,))
print(y1, y2, y3, y4)

# jacobian (gradient)
gf_x1 = grad(f)(x)
gf_x2 = grad(jit(f))(x)
gf_x3 = jit(grad(jit(f)))(x)
_, gf_x4 = jvp(f, (x,), (x_dot,))
_, gf_x5 = jvp(jit(f), (x,), (x_dot,))
gf_x6 = jit(grad(f))(x)
print(gf_x1, gf_x2, gf_x3, gf_x4, gf_x5, gf_x6)

# hessian (gradient of gradeint)
ggf_x1 = grad(grad(f))(x)
ggf_x2 = grad(grad(jit(f)))(x)
ggf_x3 = grad(jit(grad(f)))(x)
ggf_x4 = jit(grad(grad(f)))(x)
print(ggf_x1, ggf_x2, ggf_x3, ggf_x4)

_, ggf_x5 = jvp(grad(f), (x,), (x_dot,))
_, ggf_x6 = jvp(jit(grad(f)), (x,), (x_dot,))
_, ggf_x7 = jvp(jit(grad(f)), (x,), (x_dot,))
print(ggf_x5, ggf_x6, ggf_x7)



import slope


# @slope.jit.with_options(static_argnames="p")
@slope.jit
def f(x):
    # return x + 1.
    return x + g(x) + slope.tensor([1.0])


@slope.jit
def g(x):
    return x * 2


x = slope.ones(1)
# y = f(x, p=1.)
jit_output = f.jit_program(x)
print(jit_output.program)
print(jit_output.program.num_consts)
print(jit_output.consts)
# jit_output.program.save(x, dir_path="/tmp/slope_program", dry_run=True)
jit_output.program.save(x, dir_path="/tmp/slope_program")



import slope


# procedures are basically functions accessible by Tensor.procedure syntax
@slope.core.backend.procedure_set.register()
def my_relu(x):
    y = x.maximum(slope.zeros_like(x))
    return y


# To override a procedure gradient rule, define a new operator with the same name as procedure
# the impl will default to tracing procedure as a Program if not defined.

@slope.backend.operator_set.register("my_relu")
class ReLUOp(slope.core.UnaryOperator):
    def jvp(self, primals, tangents):
        (x,), (x_dot,) = primals, tangents
        y = x.my_relu()
        y_dot = x_dot * (x == y).where(slope.ones_like(x), slope.zeros_like(x))
        return [y], [y_dot]

    def T(self, cotangents, x):
        (gL_y,) = cotangents
        gL_x = (x.my_relu() > x.zeros_like()).cast(gL_y.dtype) * gL_y
        return [gL_x]

@slope.jit
def f(x):
    y = x.my_relu()
    y = y.sum()

    return y


x1 = slope.tensor([1.0, 2.0, -1.0, 0.0])
print(f"{x1=}")
y1 = f(x1)
print(f"{y1=}")

x1_dot = slope.ones_like(x1)
y1, y1_dot = slope.jvp(f, (x1,), (x1_dot,))
print(f"{y1=}, {y1_dot=}")

gf_x1 = slope.grad(f)(x1)
print(gf_x1)



import slope
import slope.nn as nn

stride = 2
padding = 2
dilation = 1
groups = 1

w_size = (64, 32, 3, 3)
x_size = (2, 32, 16, 16)

x = slope.ones(x_size)
w = slope.ones(w_size)


@slope.jit
def f(x, w):
    y = x.conv(w, groups, stride, dilation, padding)
    return y


y = f(x, w)
y, y_dot = slope.jvp(f, (x, w), (x.ones_like(), w.ones_like()))
print(f"{x.shape=}, {w.shape=}")
print(f"{y.shape=}, {y_dot.shape=}")

y, (gL_x, gL_w) = slope.value_and_grad(lambda *args: f(*args).sum(), argnums=(0, 1))(x, w)
print(f"{y.shape=},  {gL_x.shape=}, {gL_w.shape=}")

x = slope.ones(10, *x_size)
y_vmap = slope.vmap(f, in_dim=(0, None))(x, w)
print(f"{x.shape=}, {w.shape=}")
print(f"{y_vmap.shape=}")



# Copyright 2019 The IREE Authors
#
# Licensed under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

# pylint: disable=unused-variable

import gc
import logging
import numpy as np
import os
import sys
import tempfile
import traceback
import unittest

import iree.compiler
import iree.runtime

COMPILED_ADD_SCALAR = None


def compile_add_scalar():
    global COMPILED_ADD_SCALAR
    if not COMPILED_ADD_SCALAR:
        COMPILED_ADD_SCALAR = iree.compiler.compile_str(
            """
            func.func @add_scalar(%arg0: i32, %arg1: i32) -> i32 {
              %0 = arith.addi %arg0, %arg1 : i32
              return %0 : i32
            }
            """,
            target_backends=iree.compiler.core.DEFAULT_TESTING_BACKENDS,
        )
    return COMPILED_ADD_SCALAR


def create_add_scalar_module(instance):
    binary = compile_add_scalar()
    m = iree.runtime.VmModule.from_flatbuffer(instance, binary)
    return m


def create_simple_static_mul_module(instance):
    binary = iree.compiler.compile_str(
        """
        func.func @simple_mul(%arg0: tensor<4xf32>, %arg1: tensor<4xf32>) -> tensor<4xf32> {
          %0 = arith.mulf %arg0, %arg1 : tensor<4xf32>
          return %0 : tensor<4xf32>
        }
        """,
        target_backends=iree.compiler.core.DEFAULT_TESTING_BACKENDS,
    )
    m = iree.runtime.VmModule.from_flatbuffer(instance, binary)
    return m


def create_simple_dynamic_abs_module(instance):
    binary = iree.compiler.compile_str(
        """
        func.func @dynamic_abs(%arg0: tensor<?x?xf32>) -> tensor<?x?xf32> {
          %0 = math.absf %arg0 : tensor<?x?xf32>
          return %0 : tensor<?x?xf32>
        }
        """,
        target_backends=iree.compiler.DEFAULT_TESTING_BACKENDS,
    )
    m = iree.runtime.VmModule.from_flatbuffer(instance, binary)
    return m


def create_stablehlo_module(instance):
    binary = iree.compiler.compile_str(
        """
func.func @main(
  %image: tensor<28x28xf32>,
  %weights: tensor<784x10xf32>,
  %bias: tensor<1x10xf32>
) -> tensor<1x10xf32> {
  %0 = "stablehlo.reshape"(%image) : (tensor<28x28xf32>) -> tensor<1x784xf32>
  %1 = "stablehlo.dot"(%0, %weights) : (tensor<1x784xf32>, tensor<784x10xf32>) -> tensor<1x10xf32>
  %2 = "stablehlo.add"(%1, %bias) : (tensor<1x10xf32>, tensor<1x10xf32>) -> tensor<1x10xf32>
  %3 = "stablehlo.constant"() { value = dense<0.0> : tensor<1x10xf32> } : () -> tensor<1x10xf32>
  %4 = "stablehlo.maximum"(%2, %3) : (tensor<1x10xf32>, tensor<1x10xf32>) -> tensor<1x10xf32>
  "func.return"(%4): (tensor<1x10xf32>) -> ()
}
        """
        
# '''
# func.func @main (%x0: tensor<2xui64>) -> (tensor<2xui64>)
# {
#     %y_, %y = mhlo.rng_bit_generator %x, algorithm = THREE_FRY : (tensor<2xui64>) -> (tensor<2xui64>, tensor<2x2xui64>)
#     "func.return"(%y0): (tensor<2xui64>) -> ()
# }
# '''
        ,
        target_backends=iree.compiler.DEFAULT_TESTING_BACKENDS,
    )
    m = iree.runtime.VmModule.from_flatbuffer(instance, binary)
    return m


instance = iree.runtime.VmInstance()
device = iree.runtime.get_device(iree.compiler.core.DEFAULT_TESTING_DRIVER)
# device = iree.runtime.get_device("local-task")
allocator = device.allocator
hal_module = iree.runtime.create_hal_module(instance, device)


m = create_stablehlo_module(instance)
context = iree.runtime.VmContext(instance, modules=[hal_module, m])
f = m.lookup_function("main")
finv = iree.runtime.FunctionInvoker(context, device, f, tracer=None)
image =  iree.runtime.asdevicearray(device, np.ones((28,28), dtype=np.float32))
weights =  iree.runtime.asdevicearray(device,np.ones((784,10), dtype=np.float32))
bias =  iree.runtime.asdevicearray(device,np.ones((1,10), dtype=np.float32))
result = finv(image, weights, bias)
print("result:", result.to_host())

breakpoint()



from slope.tensor import Tensor
from slope.tracer_tensor import TraceTensor
import slope
from typing import (
    Sequence,
    Callable,
    NamedTuple,
    Tuple,
    List,
    Any,
    List,
    Tuple,
    Optional,
    Any,
    Union,
    Dict,
    Set,
    DefaultDict,
    Callable,
)
import numpy as np

"""
// https://www.tensorflow.org/mlir/hlo_ops#mhlogather_mhlogatherop

%result = "mhlo.scatter"(%input, %scatter_indices, %update) ({
  ^bb0(%arg0: tensor<i32>, %arg1: tensor<i32>):
    %0 = mhlo.add %arg0, %arg1 : tensor<i32>
    mhlo.return %0 : tensor<i32>
}) {
  scatter_dimension_numbers = #mhlo.scatter<
    update_window_dims = [2,3],
    inserted_window_dims = [0],
    scatter_dims_to_operand_dims = [1, 0],
    index_vector_dim = 2>,
  indices_are_sorted = false,
  unique_indices = false
} : (tensor<3x4x2xi32>, tensor<2x3x2xi64>, tensor<2x3x2x2xi32>) -> tensor<3x4x2xi32>

"""


class GatherDimensionNumbers(NamedTuple):
    """
    `index_vector_dim` is implicit, as last dimension.
    To gather scalar indices, add a trailing dimension of size 1.
    """

    offset_dims: Tuple[int, ...]
    collapsed_slice_dims: Tuple[int, ...]
    start_index_map: Tuple[int, ...]
    index_vector_dim = -1


def gather(
    operand,
    start_indices,
    dimension_numbers: GatherDimensionNumbers,
    slice_sizes,
):
    # parsed_mode = 2 # PROMISE_IN_BOUNDS
    # fill_value = None
    return operand.gather(
        operand,
        start_indices,
        dimension_numbers=dimension_numbers,
        slice_sizes=slice_sizes,
    )


def take(self, idx):
    treedef, static_idx, dynamic_idx = _split_index_for_jit(idx, self.shape)
    return _gather(self, treedef, static_idx, dynamic_idx)


def _gather(arr, treedef, static_idx, dynamic_idx):
    idx = _merge_static_and_dynamic_indices(treedef, static_idx, dynamic_idx)
    indexer = _index_to_gather(arr.shape, idx)  # shared with _scatter_update
    y = arr
    if indexer.slice_shape == ():
        return Tensor.zeros(shape=indexer.slice_shape, dtype=y.dtype)
    if indexer.gather_indices.shape != ():
        y = y.gather(indexer.gather_indices, indexer.dnums, indexer.gather_slice_shape)
    if indexer.reversed_y_dims:
        # y = y.flip(indexer.reversed_y_dims)
        raise NotImplementedError
    return y.expand_dims(indexer.newdim_dims)


class GatherDimensionNumbers(NamedTuple):
    offset_dims: Tuple[int, ...]
    collapsed_slice_dims: Tuple[int, ...]
    start_index_map: Tuple[int, ...]


class _Indexer(NamedTuple):
    slice_shape: Sequence[int]
    gather_slice_shape: Sequence[int]
    gather_indices: Tensor
    dnums: GatherDimensionNumbers
    reversed_y_dims: Sequence[int]
    newdim_dims: Sequence[int]


def _split_index_for_jit(idx, shape):
    """Splits indices into necessarily-static and dynamic parts.

    Used to pass indices into `jit`-ted function.
    """
    # Convert list indices to tuples in cases (deprecated by NumPy.)
    assert type(idx) in (int, tuple, Tensor)
    if type(idx) is int:
        idx = (idx,)
    # idx = _expand_bool_indices(idx, shape)

    leaves, treedef = slope.ad.tree_flatten(idx)
    dynamic = [None] * len(leaves)
    static = [None] * len(leaves)
    for i, x in enumerate(leaves):
        if x is Ellipsis:
            static[i] = x
        elif isinstance(x, slice):
            # slice objects aren't hashashed.
            static[i] = (x.start, x.stop, x.step)
        else:
            dynamic[i] = x
    return treedef, tuple(static), dynamic


def _merge_static_and_dynamic_indices(treedef, static_idx, dynamic_idx):
    """Recombines indices that were split by _split_index_for_jit."""
    idx = []
    for s, d in zip(static_idx, dynamic_idx):
        if d is not None:
            idx.append(d)
        elif isinstance(s, tuple):
            idx.append(slice(s[0], s[1], s[2]))
        else:
            idx.append(s)
    return treedef.unflatten(idx)


# def _canonicalize_tuple_index(arr_ndim, idx, tensor_name="tensor"):
#     """Helper to remove Ellipsis and add in the implicit trailing slice(None)."""
#     len_without_none = sum(1 for e in idx if e is not None and e is not Ellipsis)
#     if len_without_none > arr_ndim:
#         raise IndexError(
#             f"Too many indices for {tensor_name}: {len_without_none} "
#             f"non-None/Ellipsis indices for dim {arr_ndim}."
#         )
#     ellipses = (i for i, elt in enumerate(idx) if elt is Ellipsis)
#     ellipsis_index = next(ellipses, None)
#     if ellipsis_index is not None:
#         if next(ellipses, None) is not None:
#             raise IndexError(
#                 f"Multiple ellipses (...) not supported: {list(map(type, idx))}."
#             )
#         colons = (slice(None),) * (arr_ndim - len_without_none)
#         idx = idx[:ellipsis_index] + colons + idx[ellipsis_index + 1 :]
#     elif len_without_none < arr_ndim:
#         colons = (slice(None),) * (arr_ndim - len_without_none)
#         idx = tuple(idx) + colons
#     return idx


def _is_int_tensorlike(x):
    """Returns True if x is tensor-like with integer dtype, False otherwise."""
    return (
        isinstance(x, int)
        and not isinstance(x, bool)
        or getattr(x, "dtype", None) == np.integer
        or isinstance(x, (list, tuple))
        and all(_is_int_tensorlike(e) for e in x)
    )


def _normalize_index(index, dim_size):
    return (index < 0).select(index + dim_size, index)


def broadcast_tensors(*args) -> List[Tensor]:
    """Like Numpy's broadcast_tensors but doesn't return views."""
    shapes = [np.shape(arg) for arg in args]
    if not shapes or all(shapes[0] == s for s in shapes):
        return Tensor(arg)
    result_shape = _broadcast_shapes_uncached(*shapes)
    return [_expand(arg, result_shape) for arg in args]


def _expand(arr, shape) -> Tensor:
    if not isinstance(shape, tuple) and np.ndim(shape) == 0:
        shape = (shape,)
    if arr.shape == shape:
        return arr
    else:
        nlead = len(shape) - len(arr.shape)
        shape_tail = shape[nlead:]
        (diff,) = np.where(tuple((len(arr_d) != len(shape_d)) for arr_d, shape_d in zip(arr.shape, shape_tail)))
        new_dims = tuple(range(nlead)) + tuple(nlead + diff)
        kept_dims = tuple(np.delete(np.arange(len(shape)), new_dims))
        return arr.squeeze(tuple(diff)).broadcast(shape, kept_dims)


def _broadcast_shapes_uncached(*shapes):
    def _broadcast_ranks(s1, s2):
        if len(s1) > len(s2):
            s1, s2 = s2, s1
        assert len(s1) <= len(s2)
        s1_ = s2[len(s2) - len(s1) :]
        if s1_.shape == s1.shape:
            return s2
        else:
            raise ValueError

    fst, *rst = shapes
    if not rst:
        return fst

    # First check if we need only rank promotion (and not singleton-broadcasting).
    try:
        return reduce(_broadcast_ranks, rst, fst)
    except ValueError:
        pass

    # Next try singleton-broadcasting, padding out ranks using singletons.
    ndim = max(len(shape) for shape in shapes)
    shape_list = [(1,) * (ndim - len(shape)) + shape for shape in shapes]
    result_shape = _try_broadcast_shapes(shape_list)
    if result_shape is None:
        raise ValueError(f"Incompatible shapes for broadcasting: shapes={list(shapes)}")
    return result_shape


def _try_broadcast_shapes(shapes: Sequence[Tuple[int, ...]]) -> Optional[Tuple[int, ...]]:
    if len(shapes) == 1:
        return shapes[0]
    rank, *others = {len(shape) for shape in shapes}
    if others:
        return None  # must have consistent rank
    if not rank:
        return ()  # scalar case
    result_shape = []
    for ds in zip(*shapes):
        if all((d == ds[0]) for d in ds[1:]):
            result_shape.append(ds[0])
        else:
            non_1s = [d for d in ds if d != 1]
            if not non_1s:
                result_shape.append(1)
            elif all((non_1s[0] == d) for d in non_1s[1:]):
                result_shape.append(non_1s[0])
            else:
                return None
    return tuple(result_shape)


def _static_idx(idx: slice, size):
    """Helper function to compute the static slice start/limit/stride values."""
    if isinstance(size, int):
        start, stop, step = idx.indices(size)
    else:
        raise TypeError(size)

    if (step < 0 and stop >= start) or (step > 0 and start >= stop):
        return 0, 0, 1, False  # sliced to size zero

    if step > 0:
        return start, stop, step, False
    else:
        k = (start - stop - 1) % (-step)
        return stop + k + 1, start + 1, -step, True


def _index_to_gather(x_shape: Sequence[int], idx: Sequence[Any], normalize_indices: bool = True) -> _Indexer:
    # Remove ellipses and add trailing slice(None)s.
    # idx = _canonicalize_tuple_index(len(x_shape), idx)
    arr_ndim = len(x_shape)
    len_without_none = sum(1 for e in idx if e is not None and e is not Ellipsis)
    if len_without_none > arr_ndim:
        raise IndexError(
            f"Too many indices for tensor: {len_without_none} " f"non-None/Ellipsis indices for dim {arr_ndim}."
        )
    ellipses = (i for i, elt in enumerate(idx) if elt is Ellipsis)
    ellipsis_index = next(ellipses, None)
    if ellipsis_index is not None:
        if next(ellipses, None) is not None:
            raise IndexError(f"Multiple ellipses (...) not supported: {list(map(type, idx))}.")
        colons = (slice(None),) * (arr_ndim - len_without_none)
        idx = idx[:ellipsis_index] + colons + idx[ellipsis_index + 1 :]
    elif len_without_none < arr_ndim:
        colons = (slice(None),) * (arr_ndim - len_without_none)
        idx = tuple(idx) + colons
    advanced_dim_are_contiguous = False

    advanced_indexes: Optional[Sequence[Union[Tensor, np.ndarray]]] = None
    idx_advanced_dim: Sequence[int] = []
    x_advanced_dim: Optional[Sequence[int]] = None

    def _is_advanced_int_indexer(idx):
        assert isinstance(idx, tuple)
        if all(
            e is None
            or e is Ellipsis
            or isinstance(e, slice)
            or e.shape == ()
            and e.dtype in (int, np.int8, np.integer)
            for e in idx
        ):
            return False
        return all(e is None or e is Ellipsis or isinstance(e, slice) or _is_int_tensorlike(e) for e in idx)

    if _is_advanced_int_indexer(idx):
        idx_no_nones = [(i, d) for i, d in enumerate(idx) if d is not None]
        advanced_pairs = ((Tensor(e), i, j) for j, (i, e) in enumerate(idx_no_nones) if e is not None)
        if normalize_indices:
            advanced_pairs = ((_normalize_index(e, x_shape[j]), i, j) for e, i, j in advanced_pairs)
        advanced_indexes, idx_advanced_dim, x_advanced_dim = zip(*advanced_pairs)
        advanced_dim_are_contiguous = bool(np.all(np.diff(idx_advanced_dim) == 1))

    x_dim = 0  # Current dim in x.
    y_dim = 0  # Current dim in y, before collapsing. See below.
    collapsed_y_dim = 0  # Current dim in y, after collapsing.

    # Scatter dimension numbers.
    offset_dims: Sequence[int] = []
    collapsed_slice_dims: Sequence[int] = []
    start_index_map: Sequence[int] = []

    index_dtype = np.int32
    gather_indices: List[Any] = []
    gather_indices_shape: List[int] = []
    slice_shape: Sequence[int] = []

    newdim_dims: Sequence[int] = []

    reversed_y_dims: Sequence[int] = []

    gather_slice_shape: Sequence[int] = []

    for idx_pos, i in enumerate(idx):
        if advanced_indexes is not None and (
            advanced_dim_are_contiguous
            and idx_pos == idx_advanced_dim[0]
            or not advanced_dim_are_contiguous
            and idx_pos == 0
        ):
            advanced_indexes = broadcast_tensors(*advanced_indexes)
            shape = advanced_indexes[0].shape
            ndim = len(shape)

            start_dim = len(gather_indices_shape)
            gather_indices += ((a.astype(index_dtype), start_dim) for a in advanced_indexes)
            gather_indices_shape += shape

            start_index_map.extend(x_advanced_dim)
            collapsed_slice_dims.extend(x_advanced_dim)
            slice_shape.extend(shape)
            y_dim += ndim
            collapsed_y_dim += ndim

        # Per-index bookkeeping for advanced indexes.
        if idx_pos in idx_advanced_dim:
            x_dim += 1
            gather_slice_shape.append(1)
            continue

        try:
            abstract_i = TraceTensor.get_symtensor(i)
        except TypeError:
            abstract_i = None
        # Handle basic int indexes.
        if isinstance(abstract_i, (Tensor,)) and (not abstract_i.shape and abstract_i.dtype == np.integer):
            if x_shape[x_dim] == 0:
                # XLA gives error when indexing into an dim of size 0
                raise IndexError(f"index is out of bounds for dim {x_dim} with size 0")
            i = _normalize_index(i, x_shape[x_dim]) if normalize_indices else i
            i = i.astype(index_dtype)
            gather_indices.append((i, len(gather_indices_shape)))
            collapsed_slice_dims.append(x_dim)
            gather_slice_shape.append(1)
            start_index_map.append(x_dim)
            x_dim += 1
        # Handle np.newdim (None)
        elif i is None:
            slice_shape.append(1)
            newdim_dims.append(y_dim)
            y_dim += 1

        elif isinstance(i, slice):
            # Normalize the slice to use None when possible
            start, stop, step = i.start, i.stop, i.step
            if step is None or step == 1:
                step = None
            if step is None:
                if start is None or start == 0:
                    start = None
                if stop is None or (not isinstance(stop, TraceTensor) and (stop >= x_shape[x_dim])):
                    stop = None
            elif step == -1:
                step = -1

            # Handle slice(None) and slice(None, None, -1)
            if start is None and stop is None and (step is None or isinstance(step, int) and step == -1):
                if step == -1:
                    reversed_y_dims.append(collapsed_y_dim)
                slice_shape.append(x_shape[x_dim])
                gather_slice_shape.append(x_shape[x_dim])
                offset_dims.append(collapsed_y_dim)
                collapsed_y_dim += 1
                y_dim += 1
                x_dim += 1
            # Handle slice index (only static, otherwise an error is raised)
            else:
                if not all((elt == None or TraceTensor.get_symtensor(elt) is Tensor) for elt in (start, stop, step)):
                    msg = (
                        "Tensor slice indices must have static start/stop/step to be used "
                        "with NumPy indexing syntax. "
                        f"Found slice({start}, {stop}, {step}). "
                        "To index a statically sized "
                        "tensor at a dynamic position, try lax.dynamic_slice/"
                        "dynamic_update_slice (JAX does not support dynamically sized "
                        "tensors within JIT compiled functions)."
                    )
                    raise IndexError(msg)
                # if not core.is_constant_dim(x_shape[x_dim]):
                #     msg = (
                #         "Cannot use NumPy slice indexing on an tensor dimension whose "
                #         f"size is not statically known ({x_shape[x_dim]}). "
                #         "Try using lax.dynamic_slice/dynamic_update_slice"
                #     )
                #     raise IndexError(msg)
                start, limit, stride, needs_rev = _static_idx(slice(start, stop, step), x_shape[x_dim])
                if needs_rev:
                    reversed_y_dims.append(collapsed_y_dim)
                if stride == 1:
                    i = stamachine.astype(index_dtype)
                    gather_indices.append((i, len(gather_indices_shape)))
                    slice_shape.append(limit - start)
                    gather_slice_shape.append(limit - start)
                    offset_dims.append(collapsed_y_dim)
                    start_index_map.append(x_dim)
                else:
                    i = np.arange(start, limit, stride, dtype=index_dtype)
                    size = i.shape[0]
                    slice_shape.append(size)
                    gather_slice_shape.append(1)
                    gather_indices.append((i, len(gather_indices_shape)))
                    gather_indices_shape.append(size)

                    start_index_map.append(x_dim)
                    collapsed_slice_dims.append(x_dim)

                collapsed_y_dim += 1
                y_dim += 1
                x_dim += 1
        else:
            raise NotImplementedError

    if len(gather_indices) == 0:
        gather_indices_tensor = np.zeros((0,), dtype=index_dtype)
    elif len(gather_indices) == 1:
        g, _ = gather_indices[0]
        gather_indices_tensor = g.expand_dims((g.ndim,))
    else:
        last_dim = len(gather_indices_shape)
        gather_indices_shape.append(1)
        gather_indices_list = [
            g.broadcast(gather_indices_shape, tuple(range(i, i + g.ndim))) for g, i in gather_indices
        ]
        gather_indices_tensor = TraceTensor.cat(gather_indices_list, last_dim)

    return _Indexer(
        slice_shape=slice_shape,
        newdim_dims=tuple(newdim_dims),
        gather_slice_shape=gather_slice_shape,
        reversed_y_dims=reversed_y_dims,
        dnums=GatherDimensionNumbers(
            offset_dims=tuple(offset_dims),
            collapsed_slice_dims=tuple(sorted(collapsed_slice_dims)),
            start_index_map=tuple(start_index_map),
        ),
        gather_indices=gather_indices_tensor,
    )


# def _is_boolean_index(i):
#   try:
#     abstract_i = core.get_symtensor(i)
#   except TypeError:
#     abstract_i = None
#   return (isinstance(abstract_i, ShapedTensor) and issubdtype(abstract_i.dtype, bool_)
#           or isinstance(i, list) and i and all(_is_scalar(e)
#           and issubdtype(_dtype(e), np.bool_) for e in i))


# def _expand_bool_indices(idx, shape):
#     """Converts concrete bool indexes into advanced integer indexes."""
#     out = []
#     total_dims = len(shape)
#     num_ellipsis = sum(e is Ellipsis for e in idx)
#     if num_ellipsis > 1:
#         raise IndexError("an index can only have a single ellipsis ('...')")
#     elif num_ellipsis == 1:
#         total_dims = sum(e.ndim if _is_boolean_index(e) else 1 for e in idx
#                         if e is not None and e is not Ellipsis)
#     ellipsis_offset = 0
#     for dim_number, i in enumerate(idx):
#         try:
#             abstract_i = core.get_symtensor(i)
#         except TypeError:
#             abstract_i = None
#         if _is_boolean_index(i):
#             if isinstance(i, list):
#                 i = Tensor(i)
#             elif i.ndim == 0:
#                 raise TypeError("JAX tensors do not support boolean scalar indices")
#             else:
#                 i_shape = i.shape
#                 start = len(out) + ellipsis_offset
#                 expected_shape = shape[start: start + i.ndim]
#                 if i_shape != expected_shape:
#                     raise IndexError("boolean index did not match shape of indexed tensor in index "
#                                 f"{dim_number}: got {i_shape}, expected {expected_shape}")
#                 out.extend(np.where(i))
#         else:
#             out.append(i)
#         if i is Ellipsis:
#             ellipsis_offset = len(shape) - total_dims - 1
#     return tuple(out)

# tracer
"""
    def __getitem__(self, idx):
        def _is_simple_reverse_slice(idx: Any) -> bool:
            return (
                isinstance(idx, slice)
                and idx.start is idx.stop is None
                and isinstance(idx.step, int)
                and idx.step == -1
            )

        def _is_integer_index(idx: Any) -> bool:
            return isinstance(idx, (int, np.integer)) and not isinstance(
                idx, (bool, np.bool_)
            )

        def _is_valid_integer_index_for_slice(idx, size, mode):
            if size == 0:
                return False
            if _is_integer_index(idx):
                return -size <= idx < size
            try:
                shape, dtype = np.shape(idx), idx.dtype
            except:
                return False
            if shape == () and np.issubdtype(dtype, np.integer):
                True
            return False

        def _is_contiguous_slice(idx):
            return (
                isinstance(idx, slice)
                and (idx.start is None or _is_integer_index(idx.start))
                and (idx.stop is None or _is_integer_index(idx.stop))
                and (
                    idx.step is None or (_is_integer_index(idx.step) and idx.step == 1)
                )
            )

        arr = self
        # attempt to compute _rewriting_take via lax.slice(); return None if not possible.
        idx = idx if isinstance(idx, tuple) else (idx,)
        if (
            not all(isinstance(i, int) for i in arr.shape)
            or (len(idx) > arr.ndim)
            or (any(i is None for i in idx))
        ):
            raise NotImplementedError

        simple_revs = {i for i, ind in enumerate(idx) if _is_simple_reverse_slice(ind)}
        int_indices = {
            i
            for i, (ind, size) in enumerate(zip(idx, arr.shape))
            if _is_valid_integer_index_for_slice(ind, size, mode)
        }
        contiguous_slices = {
            i for i, ind in enumerate(idx) if _is_contiguous_slice(ind)
        }

        has_partial_slices = any(
            idx[i].indices(arr.shape[i]) != (0, arr.shape[i], 1)
            for i in contiguous_slices
        )
        if len(simple_revs) + len(int_indices) + len(contiguous_slices) != len(idx):
            return None

        if simple_revs:
            arr = arr.flip(tuple(simple_revs))
            idx = tuple(
                slice(None) if i in simple_revs else ind for i, ind in enumerate(idx)
            )
            contiguous_slices |= simple_revs

        if not (int_indices or has_partial_slices):
            return arr

        idx += (arr.ndim - len(idx)) * (slice(None),)
        start_indices: Sequence[TensorLike] = []
        slice_sizes: Sequence[int] = []

        for ind, size in list_zip(idx, arr.shape):
            if isinstance(ind, slice):
                start, stop, step = ind.indices(size)
                assert step == 1  # checked above
                start_indices.append(start)
                slice_sizes.append(max(0, stop - start))
        else:
            assert np.issubdtype(ind.dtype, np.integer)  # checked above
            assert np.shape(ind) == ()  # checked above
            start_indices.append(ind)
            slice_sizes.append(1)
        if len(start_indices) > 1:
            start_indices = promote_dtypes(*start_indices)
        arr = arr.slice(start_indices=start_indices, slice_sizes=slice_sizes)
        if int_indices:
            arr = arr.squeeze(tuple(int_indices))
        return arr

    def __setitem__(self, idx, val):
        raise NotImplementedError

    def gather(
        self,
        gather_indices,
        dnums,
        gather_slice_shape,
    ):
        return ops.Gather.do(self, gather_indices, dnums, gather_slice_shape)
"""
